package main

import (
	"crypto/ecdsa"
	"crypto/elliptic"
	"crypto/rand"
	"flag"
	"fmt"
	"math"
	"time"

	"github.com/ethereum/go-ethereum/crypto"
	"github.com/ethereum/go-ethereum/crypto/secp256k1"
)

var (
	THREADS int
)

func genKeyAndHashParallel(repeats int) float64 {
	done := make(chan bool)

	timeStart := time.Now()
	for i := 0; i < THREADS; i++ {
		go func() {
			for i := 0; i < repeats/THREADS; i++ {
				privKey, _ := ecdsa.GenerateKey(secp256k1.S256(), rand.Reader)
				pubkey := elliptic.Marshal(secp256k1.S256(), privKey.X, privKey.Y)
				crypto.Keccak256Hash(pubkey[1:])
			}
			done <- true
		}()
	}

	for i := 0; i < THREADS; i++ {
		<-done
	}

	duration := time.Since(timeStart)
	return duration.Seconds()
}

func genKeyAndHash(repeats int) float64 {
	timeStart := time.Now()
	for i := 0; i < repeats; i++ {
		privKey, _ := ecdsa.GenerateKey(secp256k1.S256(), rand.Reader)
		pubkey := elliptic.Marshal(secp256k1.S256(), privKey.X, privKey.Y)
		crypto.Keccak256Hash(pubkey[1:])
	}
	duration := time.Since(timeStart)

	return duration.Seconds()
}

func runBenchmark(numIDs, numRuns int, runParallel bool) (float64, float64) {
	var durationList []float64
	if runParallel == true {
		for i := 0; i < numRuns; i++ {
			durationList = append(durationList, genKeyAndHashParallel(numIDs))
		}
	} else {
		for i := 0; i < numRuns; i++ {
			durationList = append(durationList, genKeyAndHash(numIDs))
		}
	}
	sum := float64(0.0)
	for _, dur := range durationList {
		sum += dur
	}
	mean := sum / float64(numRuns)

	sd := float64(0.0)
	for _, dur := range durationList {
		sd += math.Pow(dur-mean, 2)
	}
	sd = math.Pow(sd/float64(numRuns), 0.5)

	conf := 1.95996
	dev := conf * (sd / math.Sqrt(float64(numRuns)))

	return mean, dev
}

func printPGF(numIDs int, mean, ci float64) {
	// print mean (sec) and 95% CI
	fmt.Println("(", numIDs, ",", mean, ") +- (", ci, ",", ci, ")")
}

func main() {
	runParallelPtr := flag.Bool("parallel", false, "parallel key/hash generation")
	numThreadsPtr := flag.Int("numThreads", 4, "number of threads (requrires '-parellel')")
	numRunsPtr := flag.Int("numRuns", 100, "number of repeats")
	flag.Parse()
	runParallel := *runParallelPtr
	THREADS = *numThreadsPtr
	numRuns := *numRunsPtr

	fmt.Println("=== Run Benchmark ===")
	fmt.Println("Parallel:            ", runParallel)
	fmt.Println("Number of goroutines:", THREADS)
	fmt.Println("Number of repeats:   ", numRuns)
	fmt.Println("")

	for numIDs := 5000; numIDs <= 1000000; numIDs += 5000 {
		mean, ci := runBenchmark(numIDs, numRuns, runParallel)
		printPGF(numIDs, mean, ci)
	}
}
