package gethTest

import (
	"crypto/ecdsa"
	"crypto/elliptic"
	"crypto/rand"
	"fmt"
	"math/big"

	"github.com/ethereum/go-ethereum/common"
	"github.com/ethereum/go-ethereum/crypto"
	"github.com/ethereum/go-ethereum/crypto/secp256k1"
)

type kademliaStorage struct {
	buckets [][]*idKeyTuple

	idLength        int
	numBuckets      int
	numIDsPerBucket int
}

type idKeyTuple struct {
	ID  common.Hash
	Key *ecdsa.PrivateKey
}

func getBucket(id common.Hash, numBuckets, idLen int) int {
	x := id.Big()
	div := new(big.Int)
	div.Exp(big.NewInt(2), big.NewInt(int64(idLen)), nil)
	div.Div(div, big.NewInt(int64(numBuckets)))
	x.Div(x, div)
	bucket := int(x.Int64())
	return bucket
}

// Pop returns all hash/key pairs from the bucket the specified hash falls into
func (k *kademliaStorage) Pop(targetID common.Hash) []*idKeyTuple {
	bucket := getBucket(targetID, k.numBuckets, k.idLength)
	return k.buckets[bucket]
}

// Insert puts a hash/key pair into the respective bucket
func (k *kademliaStorage) Insert(id common.Hash, key *ecdsa.PrivateKey) bool {
	bucket := getBucket(id, k.numBuckets, k.idLength)
	if len(k.buckets[bucket]) >= k.numIDsPerBucket {
		return false
	}
	newTuple := &idKeyTuple{ID: id, Key: key}
	k.buckets[bucket] = append(k.buckets[bucket], newTuple)
	return true
}

// Fill generates numBuckets*numIDsPerBucket keys to fill all buckets
func (k *kademliaStorage) Fill() {
	maxKeys := k.numBuckets * k.numIDsPerBucket
	numKeys := 0

	for numKeys < maxKeys {
		privKey := new(ecdsa.PrivateKey)
		privKeyHex := ""
		for len(privKeyHex) != 64 {
			privKey, _ = ecdsa.GenerateKey(secp256k1.S256(), rand.Reader)
			privKeyHex = fmt.Sprintf("%x", privKey.D)
		}
		pubkey := elliptic.Marshal(secp256k1.S256(), privKey.X, privKey.Y)
		pubkey = pubkey[1:]
		id := crypto.Keccak256Hash(pubkey)

		if k.Insert(id, privKey) == true {
			numKeys++
			fmt.Println("Progress:", numKeys, "/", maxKeys)
		}
	}
}

// GetAllTuples returns all stored hash/key pairs
func (k *kademliaStorage) GetAllTuples() []*idKeyTuple {
	var tuples []*idKeyTuple
	for i := range k.buckets {
		for j := range k.buckets[i] {
			tuples = append(tuples, k.buckets[i][j])
		}
	}
	return tuples
}

// NewKademliaStorage creates a new kademlia storage
func NewKademliaStorage(idLength, numBuckets, numIDsPerBucket int) *kademliaStorage {
	k := new(kademliaStorage)
	buckets := make([][]*idKeyTuple, numBuckets)
	for i := range buckets {
		buckets[i] = make([]*idKeyTuple, 0)
	}
	k.buckets = buckets
	k.idLength = idLength
	k.numBuckets = numBuckets
	k.numIDsPerBucket = numIDsPerBucket
	return k
}
